# Pfilter/Lib.pl
#
# common library functions for the filtering and webmin/gui parts of pfilter

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# list of functions in this file:

# function to return non-zero if a passed internet address is in a passed range
# sub address_in_range( $vars_ref, $address, $range, 
# 		      $error_lines_ref, $warning_lines_ref )

# function to compare two numeric ip addresses
# sub compare_ip_to_ip( $vars_ref, $address_a, $address_b )

# function to compare two address octets arrays
# sub compare_octets_to_octets( $vars_ref, $octets_a_ref, $octets_b_ref )

# function to return the non-comment part of a line as a list of fields
# sub fields_without_ruleset_comment($vars_ref, $line)

# function to convert a numeric address ip string to an octets array
# sub ip_to_octets( $ip )

# function to test if a port name matches one defined in /etc/services or
# sub known_named_port($protocol, $port)

# function to convert a mask bits size such as 24 to a ip style mask
# sub mask_to_mask_ip( $vars_ref, $mask_size )

# function to determine if an octet style address is within an address
# range specified as low and high addresses octets
# sub octet_in_octet_limits( $vars_ref, $address_ref, 
# 			     $limits_low_ref, $limits_high_ref )

# function to convert an address octets list into a numeric ip string
# sub octets_to_ip( @octets )

# function to determine if two address ranges intersect
# sub range_intersects_range( $vars_ref, $range1, $range2, $error_lines_ref, 
# 			      $warning_lines_ref )

# function to return the resolved base addresses and the mask of a range
# sub range_to_bases_mask( $vars_ref, $range, $bases_ref, $mask_ref, 
# 			   $error_lines_ref, $warning_lines_ref )

# function to return the lowest and highest addresses in a passed range
# sub range_to_low_high( $vars_ref, $range, $low_ref, $high_ref, 
# 		         $error_lines_ref, $warning_lines_ref )

# function to sort an input list and return a list of unique elements
# sub sort_uniq( @list )

#=======================================================================

# function to return non-zero if a passed internet address is in the 
# range of a passed internet address range. The range can be a single 
# address, or an address/mask. If the address resolves to more than
# one numeric address, non-zero is returned if any of them fall in the
# range in question.

sub address_in_range {
    my ( $vars_ref, $address, $range, 
	 $error_lines_ref, $warning_lines_ref ) = @_;
#    print "$0: entering address_in_range address=$address range=$range\n"
#	if ${$$vars_ref{constants}}{debug};

    # split the address in question into resolved numeric addresses
    my @ips = address_to_ips( $vars_ref, $address );
    if ( ! scalar @ips ) {
	push @$error_lines_ref, 
	"cannot determine numeric address(es) for $address";
	print "$0: exiting  address_in_range no ips for address=$address\n"
	    if ${$$vars_ref{constants}}{debug};
	return 0;
    }
    # split the range in question into resolved numeric address
    # ranges and find the absolute lowest and highest addresses
    my $range_low_address;
    my $range_high_address;
    if ( ! range_to_low_high( $vars_ref,
			      $range,
			      \$range_low_address,
			      \$range_high_address,
			      $error_lines_ref,
			      $warning_lines_ref ) ) {
	print "$0: exiting  address_in_range range_to_low_high failed for address=$address\n"
	    if ${$$vars_ref{constants}}{debug};
	return 0;
    }

    # loop through the resolved addresses and see if any of them
    # fit into the range in question
    my $result = 0;
    foreach my $ip ( @ips ) {
	$result = 1
	    if ( compare_ip_to_ip( $vars_ref, $ip, $range_low_address ) >= 0 &&
		 compare_ip_to_ip( $vars_ref, $ip, $range_high_address ) <= 0 );
    }

#    print "$0: exiting  address_in_range address=$address range=$range returning $result\n"
#	if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================

# function that when passed an address, resolves it to one numeric ip
# strings, adding a warning message if more than one resolved, or 
# adding an error message if the address did not resolve

sub address_to_ip {
    my ( $vars_ref,
	 $address,
	 $error_lines_ref,
	 $warning_lines_ref ) = @_;

    # value to be returned
    my $ip;

    # find all the numeric ip addresses that match
    my @ips = address_to_ips( $vars_ref, $address );

    # if no match at all, return an error message and a zero length ip string
    if ( ! @ips ) {
	push @$error_lines_ref,
	"cannot resolve address $address to a numeric ip address";
	$ip = "";
    }
    # if too many matches, use the first one and output a warning message
    elsif ( scalar @ips > 1 ) {
	$ip = $ips[0];
	push @$warning_lines_ref,
	"address $address resolved to more than one ip, using $ip";
    }
    # otherwise, use the one that was resolved
    else {
	$ip = $ips[0];
    }
    return $ip;
}

#=======================================================================

# function that when passed an address, resolves it to one or more 
# numeric ip strings and returns an array of those strings

sub address_to_ips {
    my ($vars_ref, $address ) = @_;

    my @ips = ();

    # if the address is already a numeric ip string, return that,
    # otherwise, a local name or dns name, convert to numeric ip
    if ( valid_ip( $vars_ref, $address ) ) {
	push @ips, $address;
    } else {
	print "$0: in       address_to_ips resolving address $address\n"
	    if ${$$vars_ref{constants}}{debug};
	my ( $name, $aliases, $addrtype, $length, @addrs) = 
	    gethostbyname $address;
	foreach my $raw ( @addrs ) {
	    my ( $b1, $b2, $b3, $b4 ) = unpack('C4', $raw);
	    push @ips, "$b1.$b2.$b3.$b4";
	}
        if ( ${$$vars_ref{constants}}{debug} ) {
            my $string = join( ',', @ips );
            print "$0: in       address_to_ips resolved address $address to ips ($string)\n";
	}
    }
#    if ( ${$$vars_ref{constants}}{debug} ) {
#        my $string = join( ',', @ips );
#        print "$0: exiting  address_to_ips address=$address returning ips=($string)\n";
#    }
    return @ips;
}

#=======================================================================

# function to compare two numeric ip addresses, returning -1 if the 
# first address is lower, 0 if they are the same, or 1 if the first
# address is higher.

sub compare_ip_to_ip {
    my ( $vars_ref, $address_a, $address_b ) = @_;

    my @octets_a = ip_to_octets( $address_a );
    my @octets_b = ip_to_octets( $address_b );

    return compare_octets_to_octets( $vars_ref,
				     \@octets_a,
				     \@octets_b );
}

#=======================================================================

# function to compare two address octets arrays, returning -1 if the 
# first octet has a lower ip address, 0 if they are the same, or
# 1 if the first octet has a higher ip address. this function does not
# check to see if either octets array is valid, it just compares the
# octets arrays until a difference is found.

sub compare_octets_to_octets {
    my ( $vars_ref, $octets_a_ref, $octets_b_ref ) = @_;
    my $index = 0;
    my $result = 0;

    while ( $index < $#$octets_a_ref && $index < $#$octets_b_ref ) {
	if ( $$octets_a_ref[$index] < $$octets_b_ref[$index] ) {
	    $result = -1;
	    last;
	}
	if ( $$octets_a_ref[$index] > $$octets_b_ref[$index] ) {
	    $result = 1;
	    last;
	}
	$index++;
    }
#    if ( ${$$vars_ref{constants}}{debug} ) {
#        my $string_a = join( ',', @$octets_a_ref );
#        my $string_b = join( ',', @$octets_b_ref );
#        print "$0: exiting  compare_octets_to_octets ($string_a) ($string_b) returning $result\n";
#    }
    return $result;
}

#=======================================================================

# function to return the non-comment part of a line as a list of 
# non-blank text strinqs before any comment portion. This should 
# only be used on ruleset lines, not on configuration file lines.
#
# Note that this does not count %non-blank-character(s)% text chunks
# as the start of a comment, whether they are surrounded by white 
# space or whether they are embedded in the middle of a string, since 
# constants and defines that haven't been expanded yet look like that.
#
# Note also that this does not count a conditional trigger such as
# <%if EXPRESSION> as being a comment, nor does it count directives
# such as %macro at the beginning of lines as being a comment

sub fields_without_ruleset_comment {
    my ($vars_ref, $line) = @_;
#    my $debug = ${$$vars_ref{constants}}{debug};
#    print "$0: entering fields_without_ruleset_comment line=<$line>\n" 
#        if $debug;
    my @wo_comment_fields = ();
    return @wo_comment_fields 
	if ( $line eq "" );
    my @line_fields = split( ' ', $line);
    push @wo_comment_fields, shift @line_fields
	if $line =~ /^\s*\%/ &&
	    $line_fields[0] =~
            /^\%(constant|define|else|endloop|endmacro|error|if|ifdef|ifndef|loop|macro|warning)$/;
    my $found_comment = 0;
    while ( ! $found_comment && scalar @line_fields ) {
	$line_field = shift @line_fields;
	if ( $line_field =~ /^\%/ &&          # if the field starts with %
	     $line_field !~ /^\%[^\%]+\%/ &&  # and it isn't a define/constant
	     $line_field !~ /^\%(if|ifdef|ifndef)$/ ) { # and it isn't a conditional
	    $found_comment = 1;
	} else {
	    push @wo_comment_fields, $line_field; # otherwise add to non-comments
	}
    }
#    if ( $debug ) {
#        my $string = join( ",", @wo_comment_fields );
#        print "$0: exiting  fields_without_ruleset_comment returning ($string)\n";
#    }
    return @wo_comment_fields;
}

#=======================================================================

# finds the executable path of a program at given path if the passed
# argument includes a leading / or on the PATH

sub find_executable {
    my $name = shift;
    return $name if $name =~ m#^/# && -x $name;
    $name =~ s,^.*/,,;
    foreach my $directory ( split(/:/, $ENV{PATH}) ) {
	return "$directory/$name" if -x "$directory/$name";
    }
    return undef;
}

#=======================================================================

# find the kernel name, returns the name

sub find_kernel_name {
    my $kernel_name = `uname -s`;
    chomp $kernel_name;
    return $kernel_name;
}

#=======================================================================

# find the kernel version numbers, returns the major and minor numbers

sub find_kernel_version {
    $_ = `uname -r`;
    my ($kernel_major, $kernel_minor) = /^(\d+)\.(\d+)/;
    return ( $kernel_major, $kernel_minor );
}

#=======================================================================

# function initialize the vars structure

sub init_vars {

    # Varaibles that are available during macro expansion as %CONSTANT's.
    # These cannot be changed by any ruleset %DEFINE or %CONSTANT statements.
    # Some of these are filled in when converting from pfilter.in to
    # pfilter by executing this command:
    # make pfilter

    my %constants = (
        action           => "restart",
	date             => `date`,
	debug            => 0,
        domainname       => `domainname`,
        dumpconf         => 0,                       # secret debug flag
        dumpservices     => 0,                       # secret debug flag
        exec             => 1,
        forced_mode      => "",
        hostname         => `hostname`,
        logging          => "none",
        mode             => "",
        nocond           => 0,                       # secret debug flag
        noconst          => 0,                       # secret debug flag
        nomacro          => 0,                       # secret debug flag
        nooptpath        => 0,
        noout            => 0,
        noprefix         => 0,                       # secret debug flag
        nosuffix         => 0,                       # secret debug flag
        uname            => `uname`,
        verbose          => 0,
        warnings         => 0
    );
    $constants{date} =~ s/  / /;  # get rid of souble spaces from some dates
    foreach my $key ( keys %constants ) { # get rid of new-line chars on end
        chomp $constants{$key};           # of all constants because some
    }                                     # are command outputsxw
    $constants{domainname} = "" if $constants{domainname} eq "(none)";

    # Variables that are available during macro expansion as %DEFINE's.
    # These can be changed by ruleset %DEFINE or %CONSTANT statements.
    # Some of these are filled in when converting from pfilter.in to
    # pfilter by executing this command:
    # make pfilter
    
    my %defines = (
        log_burst       => "50",
        log_level       => "DEBUG",
        log_rate        => "2/m",
        nowayed         => "REJECTED",
        nowayit         => "REJECT"
    );

    # icmp packet type names
    
    my %icmptypes = (
        "address-mask-reply"	=> 1,
        "address-mask-request"	=> 1,
        "communication-prohibited"	=> 1,
        "destination-unreachable"	=> 1,
        "echo-reply"		=> 1,
        "echo-request"		=> 1,
        "fragmentation-needed"	=> 1,
        "host-precedence-violation"	=> 1,
        "host-prohibited"		=> 1,
        "host-redirect"		=> 1,
        "host-unknown"		=> 1,
        "host-unreachable"		=> 1,
        "ip-header-bad"		=> 1,
        "network-prohibited"	=> 1,
        "network-redirect"		=> 1,
        "network-unknown"		=> 1,
        "network-unreachable"	=> 1,
        "parameter-problem"		=> 1,
        "ping"			=> 1,
        "pong"			=> 1,
        "port-unreachable"		=> 1,
        "precedence-cutoff"		=> 1,
        "protocol-unreachable"	=> 1,
        "redirect"			=> 1,
        "required-option-missing"	=> 1,
        "router-advertisement"	=> 1,
        "router-solicitation"	=> 1,
        "source-quench"		=> 1,
        "source-route-failed"	=> 1,
        "time-exceeded"		=> 1,
        "timestamp-reply"		=> 1,
        "timestamp-request"		=> 1,
        "TOS-host-redirect"		=> 1,
        "TOS-host-unreachable"	=> 1,
        "TOS-network-redirect"	=> 1,
        "TOS-network-unreachable"	=> 1,
        "ttl-exceeded"		=> 1,
        "ttl-zero-during-reassembly"=> 1,
        "ttl-zero-during-transit"	=> 1
    );
    
    # language keywords not allowed as macro or variable names
    
    my %keywords = (
		    alias	=> 1,
		    debug	=> 1,
		    drop	=> 1,
		    extif	=> 1,
		    filtered    => 1,
		    from	=> 1,
		    intif	=> 1,
		    on	        => 1,
		    onto	=> 1,
		    open	=> 1,
		    protected	=> 1,
		    reject	=> 1,
		    to		=> 1,
		    trusted	=> 1,
		    unfiltered	=> 1,
		    unprotected	=> 1,
		    verbose	=> 1
    );
    
    # macros defined in both the configuration file and the ruleset files
    
    my %macros;
    
    # all program variables that more than one portion of the code might
    # need are under this hash.  Some of these are filled in when 
    # converting from pfilter.in to pfilter by executing this command:
    # make pfilter
    
    my %vars = (
        constants       => \%constants,
        defines         => \%defines,
        icmptypes       => \%icmptypes,
        keywords        => \%keywords,
        macros          => \%macros
    );

    return \%vars;
}

#=======================================================================

# function to convert a numeric address ip string to an octets array

sub ip_to_octets {
    my( $ip ) = @_;

    my @octets = split /\./, $ip;
    return @octets;
}

#=======================================================================

# function to test if a port name matches one defined in /etc/services or
# wherever, and if so, return the list of valid protocols for it
=pod
sub known_named_port {
    my $port = shift;
    my @protocols;

#    print "$0: entering known_named_port port=$port\n" if ${$$vars_ref{constants}}{debug};
    my @protocol_matches = grep( /^$protocol\s/, @known_services);
    foreach (@protocol_matches) {
	my @fields = split;
	return 1 if $fields[1] eq $port;
    }
    if ( $protocol eq "icmp" ) {
	return 1 if ( $port >= 0 ) && ( $port <= 255 );
	return 1 if grep ( /^$port$/, @icmptypes );
    }
    return 0;
}

#=======================================================================

# function to test if a protocol/port is valid

sub known_protocol_port {
    my ($protocol, $port) = @_;
#    print "$0: entering known_port protocol=$protocol port=$port\n" if ${$$vars_ref{constants}}{debug};
    my @protocol_matches = grep( /^$protocol\s/, @known_services);
    foreach (@protocol_matches) {
	my @fields = split;
	return 1 if $fields[1] eq $port;
    }
    if ( $protocol eq "icmp" ) {
	return 1 if ( $port >= 0 ) && ( $port <= 255 );
	return 1 if grep ( /^$port$/, @icmptypes );
    }
    return 0;
}
=cut
#=======================================================================

# function to convert a mask bits size such as 24 to a ip style mask
# such as 255.255.255.0, note that minimal limits checking is done

# Passed input parameters are:
#   reference to hash of useful flags/values
#   mask size bits string
# Returned parameters are:
#   mask ip string

sub mask_to_mask_ip {
    my ( $vars_ref, $mask_size ) = @_;

    use integer;

    my $result;

    if ( $mask_size =~ /\./ ) {
	$result = $mask_size;
    } else {
	my @octets = ();
	my $mask_bits_left = $mask_size;
	while ( $mask_bits_left >= 8 && scalar @octets < 6 ) {
	    push @octets, "255";
	    $mask_bits_left -= 8;
	}
	push @octets, ( 255 << ( 7 - ( $mask_bits_left % 8 ) ) ) & 255
	    if $mask_bits_left > 0;
	push @octets, "0"
	    while scalar @octets < 4;
	$result = octets_to_ip( @octets );
    }

    print "$0: exiting  mask_to_mask_ip mask_size=$mask_size returning $result\n"
	if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================

# function to determine if an octet style address is within an address
# range specified as low and high addresses octets

# Passed input parameters are:
#   reference to hash of useful flags/values
#   reference to address in question octet
#   reference to range lowest address octet
#   reference to range highest address octet
# Returned parameters are:
#   1 or 0 if the address lies within the range

sub octet_in_octet_limits {
    my ( $vars_ref, $address_ref, 
	 $limits_low_ref, $limits_high_ref ) = @_;
    
    my $result = 
	( compare_octets_to_octets( $vars_ref,
				    $address_ref,
				    $limits_low_ref ) >= 0 )
	&&
	( compare_octets_to_octets( $vars_ref,
				    $address_ref,
				    $limits_high_ref ) <= 0 );

#    print "$0: exiting  octet_in_octet_limits address=<$$address_ref> low=<$$limits_low_ref> high=<$$limits_high_ref> returning $result\n"
#	if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================

# function to convert an address octets list into a numeric ip string

sub octets_to_ip {
    my( @octets ) = @_;

#    return join('.', @octets);
    return "$octets[0].$octets[1].$octets[2].$octets[3]";
}

#=======================================================================

# prints a dump of a hash, passed the hash name and the hash reference

sub print_hash {
    my( $leading_spaces, $name, $hashref ) = @_;
    print "$0: $leading_spaces$name ->\n";
    foreach my $key ( sort keys %$hashref ) {
	my $value = $$hashref{$key};
	if (ref($value) eq "HASH") {
	    print_hash(  "$leading_spaces    ", $key, $value );
	} elsif (ref($value) eq "ARRAY") {
	    my $string = join(',', @$value);
	    print "$0: $leading_spaces    $key => ($string)\n";
	} elsif (ref($value) eq "SCALAR") {
	    print "$0: $leading_spaces    $key is a scalar ref\n";
	    print "$0: $leading_spaces    $key => $$value\n";
	} else {
	    $value = "undef" unless defined $value;
	    print "$0: $leading_spaces    $key => <$value>\n";
	}
    }
}

#=======================================================================

# add a protocol/port value to a protocols_ports hash

sub protocols_ports_add_protocol_port {
    my($vars_ref, $destination_protocols_ports_ref, $protocol, $port) = @_;
#    if ( ${$$vars_ref{constants}}{debug} ) {
#	print( "$0: entering protocols_ports_add_protocol_port adding <$protocol>/<$port>\n" );
#	print_hash( "", "to",
#		    $destination_protocols_ports_ref);
#    }

    # if the protocol is not in the destination hash yet,
    # add it with an empty associated ports list
    unless ( exists $$destination_protocols_ports_ref{$protocol} ) {
	my %ports_hash = ();
	$$destination_protocols_ports_ref{$protocol} = \%ports_hash;
    }

    # add the port to the ports list for that protocol,
    my $destination_ports_ref = $$destination_protocols_ports_ref{$protocol};
    $$destination_ports_ref{$port} = 1;
#    if ( ${$$vars_ref{constants}}{debug} ) {
#	print_hash( "", "yields destination",
#		    $destination_protocols_ports_ref);
#    }
}

#=======================================================================

# add protocols/ports values from a protocols_ports hash to another
# protocols_ports hash

sub protocols_ports_add_protocols_ports {
    my($vars_ref, $destination_protocols_ports_ref,
       $source_protocols_ports_ref) = @_;
#    if ( ${$$vars_ref{constants}}{debug} ) {
#	print_hash( "", "in protocols_ports_add_protocols_ports adding",
#		    $source_protocols_ports_ref );
#	print_hash( "", "to",
#		    $destination_protocols_ports_ref);
#    }

    # for each protocol in the source hash, ...
    foreach my $protocol ( keys %$source_protocols_ports_ref ) {

	# if the protocol is not in the destination hash yet,
	# add it with an empty associated ports list
	unless ( exists $$destination_protocols_ports_ref{$protocol} ) {
	    my %ports_hash = ();
	    $$destination_protocols_ports_ref{$protocol} = \%ports_hash;
	}

	# for each port in that protocol in the source hash, ...
	my $source_ports_ref      = $$source_protocols_ports_ref{$protocol};
	my $destination_ports_ref = $$destination_protocols_ports_ref{$protocol};
	foreach my $port ( keys %$source_ports_ref ) {
	    #print "$0: exiting  ppapp port=$port\n" if ${$$vars_ref{constants}}{debug};

	    # add the port to the ports list for that protocol,
	    $$destination_ports_ref{$port} = 1;
	}
    }
#    if ( ${$$vars_ref{constants}}{debug} ) {
#	print_hash( "", "yields destination",
#		    $destination_protocols_ports_ref);
#    }
}

#=======================================================================

# function to determine if two address ranges intersect. This function
# resolves non-numeric ip ranges to numeric ip ranges.
#
# Passed input parameters are:
#   reference to hash of useful flags/values
#   first range string
#   second range string
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be possibly added to
# Returned parameters are:
#   1 or 0 if the ranges intersect or not

sub range_intersects_range {
    my ( $vars_ref, $range1, $range2, $error_lines_ref, 
	 $warning_lines_ref ) = @_;
#    print "$0: entering range_intersects_range range1=$range1 range2=$range2\n"
#	if ${$$vars_ref{constants}}{debug};

    # resolve ranges to numeric ip ranges
    my $resolved_range1;
    my $resolved_range2;
    my $result = 1;
    $result = 0
	if ! range_to_ip_range ( $vars_ref,
				 $range1, 
				 \$resolved_range1,
				 $error_lines_ref,
				 $warning_lines_ref );
    $result = 0
	if ! range_to_ip_range ( $vars_ref,
				 $range2, 
				 \$resolved_range2,
				 $error_lines_ref,
				 $warning_lines_ref );
    
    # if one or both of the input ranges was not valid or resolvable,
    # we're done and return false
    if ( ! $result ) {
	print "$0: exiting  range_intersects_range range1=$range1 range2=$range2 returning 0 because invalid range(s)\n"
	    if ${$$vars_ref{constants}}{debug};
        return 0;
    }

    # convert the numeric ranges into low and high ip addresses
    my $range1_low;
    my $range1_high;
    $result = 0
        if ! range_to_low_high( $vars_ref,
				$resolved_range1, 
				\$range1_low,
				\$range1_high,
				$error_lines_ref,
				$warning_lines_ref );
    my $range2_low;
    my $range2_high;
    $result = 0
        if ! range_to_low_high( $vars_ref,
				$resolved_range2, 
				\$range2_low,
				\$range2_high,
				$error_lines_ref,
				$warning_lines_ref );

    # convert the numeric low and high ip addresses to octets
    my @range1_low_oct = ip_to_octets( $range1_low );
    my @range1_high_oct = ip_to_octets( $range1_high );
    my @range2_low_oct = ip_to_octets( $range2_low );
    my @range2_high_oct = ip_to_octets( $range2_high );

    # if either of the resolved ranges is simply a single address,
    # use simpler logic to save time
    if ( $range1_low eq $range1_high ) {
	$result = octet_in_octet_limits( $vars_ref,
					 \@range1_low_oct,
					 \@range2_low_oct,
					 \@range2_high_oct );
    } elsif ( $range2_low eq $range2_high ) {
	$result = octet_in_octet_limits( $vars_ref,
					 \@range2_low_oct,
					 \@range1_low_oct,
					 \@range1_high_oct );
    }

    # if both ranges are actual ranges, use the more complicated logic
    else {
	$result = 
	    octet_in_octet_limits( $vars_ref,
				   \@range1_low_oct,
				   \@range2_low_oct,
				   \@range2_high_oct )
		||
	    octet_in_octet_limits( $vars_ref,
				   \@range1_high_oct,
				   \@range2_low_oct,
				   \@range2_high_oct )
		||
	    octet_in_octet_limits( $vars_ref,
				   \@range2_low_oct,
				   \@range1_low_oct,
				   \@range1_high_oct )
		||
	    octet_in_octet_limits( $vars_ref,
				   \@range2_high_oct,
				   \@range1_low_oct,
				   \@range1_high_oct );
    }

    print "$0: exiting  range_intersects_range range1=$range1 range2=$range2 returning $result\n"
	if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================

# function to return the resolved base addresses and the mask of 
# an address range which is passed as a string composed of the 
# unresolved base address and the address mask. This function checks
# the format of the input range.  Non-zero or zero is returned for
# success or failure.
#
# Passed input parameters are:
#   reference to hash of useful flags/values
#   range string such as "host.dns.name/255.255.255.0" or "host.dns.name/24"
#   reference to array to put resolved base addresses into
#   reference to scalar into which to put the split off mask value
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be possibly added to
# Returned parameters are:
#   1 or 0 if a valid range or invalid range is passed

sub range_to_bases_mask {
    my ( $vars_ref, $range, $bases_ref, $mask_ref, 
	 $error_lines_ref, $warning_lines_ref ) = @_;
#    print "$0: entering range_to_bases_mask range=$range\n"
#	if ${$$vars_ref{constants}}{debug};
    @$bases_ref = ();
    my $base;

    # split off the mask, or set as 255.255.255.255 if none there,
    # converting single numeric masks such as 24 to 255.255.255.0
    if ( $range =~ /\// ) {
	( $base, $$mask_ref ) = split /\//, $range, 2;
	if ( ! valid_mask( $vars_ref, $$mask_ref )) {
	    push @$error_lines_ref, "invalid mask in $range";
	    $$mask_ref="";
	    print "$0: exiting  range_to_bases_mask valid_mask returned false\n"
		if ${$$vars_ref{constants}}{debug};
	    return 0;
        } elsif ( $$mask_ref !~ /\./ ) {
	    $$mask_ref = mask_to_mask_ip( $vars_ref, $$mask_ref );
	}
    } else {
	$base = $range;
	$$mask_ref = "255.255.255.255";
    }

    # find the numeric address(es) that match this numeric/dns address
    @$bases_ref = address_to_ips( $vars_ref, $base );
    if ( ! @$bases_ref ) {
	push @$error_lines_ref, "invalid address/range $range";
	print "$0: exiting  range_to_bases_mask invalid address/range $range\n"
	    if ${$$vars_ref{constants}}{debug};
	return 0;
    }

#    if ( ${$$vars_ref{constants}}{debug} ) {
#        my $string = join( ',', @$bases_ref );
#        print "$0: exiting  range_to_bases_mask range=$range returning bases=($string) mask=$$mask_ref returning 1\n";
#    }
    return 1;
}

#=======================================================================

# function that when passed an address range returns a resolved numeric
# address range
#
# Passed input parameters are:
#   reference to hash of useful flags/values
#   range string such as "host.dns.name/255.255.255.0" or "1.2.3.4/16"
#   reference to scalar into which to put the resolved numeric range
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be possibly added to
# Returned parameters are:
#   1 or 0 if a valid range or invalid range is passed

sub range_to_ip_range {
    my ( $vars_ref, $input_range, $resolved_range_ref,
	 $error_lines_ref, $warning_lines_ref ) = @_;
#    print "$0: entering range_to_ip_range input_range=$input_range\n"
#	if ${$$vars_ref{constants}}{debug};

    # if no address mask supplied (the input was really just
    # a single address rather than a range) supply a mask
    my $range = $input_range;
    $range = "$range/255.255.255.255"
	if $range !~ /\//;

    # resolve the input range to a numeric ip low and high address,
    # then combine the low numeric ip address with the address mask
    # portion of input range
    my $low_ip;
    my $high_ip;
    my $result = range_to_low_high( $vars_ref,
				    $range,
				    \$low_ip,
				    \$high_ip,
				    $error_lines_ref,
				    $warning_lines_ref );
    # this substitutes the low numeric ip for the input base address
    $range =~ s/^[^\/]*/$low_ip/
	if $result;

    $$resolved_range_ref = $range;
    print "$0: exiting  range_to_ip_range input=$input_range resolved=$$resolved_range_ref returning $result\n"
	if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================

# function to return the lowest and highest addresses in a passed range.
# this function resolves domain addresses and checks the format of the
# input range. The low and high addresses are returned into scalars 
# whose references were given as input arguments. non-zero or zero is
# returned for success or failure.
#
# Passed input parameters are:
#   reference to hash of useful flags/values
#   range string such as "host.dns.name/255.255.255.0"
#   reference to array to put resolved base addresses into
#   reference to scalar into which to put the split off mask value
#   reference to array of error lines to be possibly added to
#   reference to array of warning message lines to be possibly added to
# Returned parameters are:
#   1 or 0 if a valid range or invalid range is passed

sub range_to_low_high {
    my ( $vars_ref, $range, $low_ref, $high_ref, 
	 $error_lines_ref, $warning_lines_ref ) = @_;
#    print "$0: entering range_to_low_high range=$range\n"
#	if ${$$vars_ref{constants}}{debug};
    my @low_oct = ( 255, 255, 255, 255 );
    my @high_oct = ( 0, 0, 0, 0 );

    # find the resolved base addresses and the mask
    my @resolved_base_addresses;
    my $mask;
    if ( ! range_to_bases_mask( $vars_ref,
				$range,
				\@resolved_base_addresses,
				\$mask,
				$error_lines_ref,
				$warning_lines_ref ) ) {
	print "$0: exiting  range_to_low_high range=$range range_to_bases_mask failed\n"
	    if ${$$vars_ref{constants}}{debug};
	return 0;
    }

    # loop through the base numeric addresses
    if ( ! scalar @resolved_base_addresses ) {
	print "$0: exiting  range_to_low_high range=$range empty resolved ips\n"
	    if ${$$vars_ref{constants}}{debug};
	return 0;
    }
    foreach my $base_address ( @resolved_base_addresses ) {
	# split this base address into octets and
	# update the low or high values if needed
	my @base_oct = ip_to_octets( $base_address );
        if ( ${$$vars_ref{constants}}{debug} ) {
            my $low_string = join( ',', @low_oct );
            my $high_string = join( ',', @high_oct );
            my $base_string = join( ',', @base_oct );
#            print "$0: in       range_to_low_high low=($low_string) high=($high_string) base=($base_string)\n";
        }
        my $compare_result;
        $compare_result = compare_octets_to_octets( $vars_ref,
                                                    \@base_oct,
                                                    \@low_oct );
#        print "$0: in       range_to_low_high compare_result=$compare_result\n"
#	    if ${$$vars_ref{constants}}{debug};
        if ( $compare_result < 0 ) {
#	    if ( ${$$vars_ref{constants}}{debug} ) {
#                my $string_base = join( '.', @base_oct );
#                my $string_low = join( '.', @low_oct );
#                print "$0: in       range_to_low_high replacing low $string_low with $string_base\n";
#            }
	    @low_oct = ( @base_oct );
        }
        $compare_result = compare_octets_to_octets( $vars_ref,
                                                    \@base_oct,
                                                    \@high_oct );
#        print "$0: in       range_to_low_high compare_result=$compare_result\n"
#	    if ${$$vars_ref{constants}}{debug};
        if ( $compare_result > 0 ) {
#	    if ( ${$$vars_ref{constants}}{debug} ) {
#                my $string_base = join( '.', @base_oct );
#                my $string_high = join( '.', @high_oct );
#                print "$0: in       range_to_low_high replacing high $string_high with $string_base\n";
#            }
	    @high_oct = ( @base_oct );
        }
    }

    # now if we had an address mask we have to account for 
    # that by masking off and on the mask on the low and
    # high base addresses
    if ( $mask ne "255.255.255.255" ) {
	my @mask_oct = ip_to_octets( $mask );
        if ( ${$$vars_ref{constants}}{debug} ) {
            my $low_string = join( ',', @low_oct );
            my $high_string = join( ',', @high_oct );
            my $mask_string = join( ',', @mask_oct );
#            print "$0: in       range_to_low_high low=($low_string) high=($high_string) mask=($mask_string)\n"
#		if ${$$vars_ref{constants}}{debug};
        }
	foreach my $index ( 0 .. 3 ) {
	    use integer;
#            print "$0: in       range_to_low_high index=$index low_byte=$low_oct[$index] high_byte=$high_oct[$index] mask_byte=$mask_oct[$index]\n"
#		if ${$$vars_ref{constants}}{debug};
            my $int_mask_value = 0+$mask_oct[$index];
            my $or_value = 0+255 & ~$int_mask_value;
#            print "$0: in       range_to_low_high         int_mask_value=$int_mask_value or_value=$or_value\n"
#		if ${$$vars_ref{constants}}{debug};
	    $low_oct[$index] = 0+$low_oct[$index] & $int_mask_value;
            $high_oct[$index] = 0+$high_oct[$index] | 0+$or_value;
#            print "$0: in       range_to_low_high         low_byte=$low_oct[$index] high_byte=$high_oct[$index]\n"
#		if ${$$vars_ref{constants}}{debug};
	}
        if ( ${$$vars_ref{constants}}{debug} ) {
            my $low_string = join( ',', @low_oct );
            my $high_string = join( ',', @high_oct );
#            print "$0: in       range_to_low_high masked low=($low_string) high=($high_string)\n"
#		if ${$$vars_ref{constants}}{debug};
        }
    }

    $$low_ref = octets_to_ip( @low_oct );
    $$high_ref = octets_to_ip( @high_oct );
    print "$0: exiting  range_to_low_high range=$range returning low=$$low_ref high=$$high_ref\n"
	if ${$$vars_ref{constants}}{debug};
    return 1;
}

#=======================================================================

# die if a required executable is not found or return proper path for it

sub required_executable {
    my $program = shift;
    my $path = find_executable($program);
    if ( $path ) {
	return $path;
    } else {
	my $name = $program;
	$name =~ s,^.*/,,;
	die "Cannot find program $name at $program or in PATH\n";
    }
}

#=======================================================================

# return a sorted, unique lines only version of a passed list

sub sort_uniq {
    my ( @input_list ) = @_;
    my %seen;
    foreach my $value ( @input_list ) {
	$seen{$value}++;
    }
    my @result = sort ( keys %seen );
    return @result;
}

#=======================================================================

# return non-zero if a passed ip address is valid (can include mask).
# this function also returns a list of resolved numeric addresses
# into an array whose reference was passed as an input argument.

sub valid_address {
    my ( $vars_ref, $address, $resolved_ips_ref ) = @_;
#    print "$0: entering valid_address address=$address\n" 
#        if ${$$vars_ref{constants}}{debug};
    my ($base, $mask);
    @$resolved_ips_ref = ();

    # if the address includes a mask split off and check the mask
    if ( $address =~ /\// ) {
	my @fields = split( /\//, $address );
	return 0 if scalar @fields != 2;
	$base = $fields[0];
	$mask = $fields[1];
	return 0 if ! valid_mask( $vars_ref, $mask );
    } else {
	$base = $address;
	$mask = "";
    }
    # find the list of ip addresses that match this base address
    @$resolved_ips_ref = address_to_ips( $vars_ref, $base );
#    print "$0: exiting  valid_address address=<$address> base=<$base> mask=<$mask> resolved_ips=<@$resolved_ips_ref>\n"
#	if ${$$vars_ref{constants}}{debug};
    return ( @$resolved_ips_ref ) ? 1 : 0;
}

#=======================================================================

# return non-zero if a passed ip address is value N.N.N.N form

sub valid_ip {
    my ($vars_ref, $ip) = @_;
#    print "$0: entering valid_ip ip=$ip\n"
#        if ${$$vars_ref{constants}}{debug};
    my $result = 0;
    if ( $ip =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/ ) {
	if ( $1 <= 255 && $2 <= 255 && $3 <= 255 && $4 <= 255 )  {
	    $result = 1;
	}
    }
#    print "$0: exiting  valid_ip ip=$ip returning $result\n"
#        if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================

# return non-zero if a passed mask value is valid

sub valid_mask {
    my ( $vars_ref, $mask ) = @_;
#    print "$0: entering valid_mask mask=$mask\n" if ${$$vars_ref{constants}}{debug};
    my $result = ( 
		   ( $mask =~ /^(\d+)$/ 
		     && $1 <= 32 )
		   ||
		   ( $mask =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/
		     && $1 <= 255 
		     && $2 <= 255 
		     && $3 <= 255 
		     && $4 <= 255 ) 
		                    )
	? 1 : 0;
#    print "$0: exiting  valid_mask mask=$mask returning $result\n"
#        if ${$$vars_ref{constants}}{debug};
    return $result;
}

#=======================================================================

# return non-zero if a passed protocol/port is valid

sub valid_port {
    my ($vars_ref, $protocol, $port) = @_;
#    print "$0: entering valid_port protocol=$protocol port=$port\n" if ${$$vars_ref{constants}}{debug};
    return 1 if $port =~ /^\d+$/ && $port <= 65535;
    return 1 if known_protocol_port( $protocol, $port );
    return 0;
}

#=======================================================================

# return non-zero if a passed ip address is an rfc 1918 address

sub valid_rfc1918_ip {
    my ($vars_ref, $ip) = @_;
    print "$0: entering valid_rfc1918_ip ip=$ip\n" if ${$$vars_ref{constants}}{debug};
    return 0 unless valid_ip ( $vars_ref, $ip );
    my @fields = $ip =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;
    return 0 if @fields != 4;
    return 1 if $fields[0] == 10;
    return 1 if $fields[0] == 172 && $fields[1] >= 16 && $fields[1] <= 31;
    return 1 if $fields[0] == 192 && $fields[1] == 168;
    return 0;
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;
