# Pfilter/Parse.pm
#
# common library functions for parsing text

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# functions in this file:

# function to parse a constant or define line
# sub parse_constant_define_from_line( $vars_ref, $line, 
#                                      $fields_without_comment_ref,
# 				       $error_lines_ref, $warning_lines_ref, $where )

# function to parse constants, defines, and macros definitions from 
# a list of text lines
# sub parse_constant_define_macro_definitions_from_lines( $vars_ref, 
# 							  $lines_ref, 
# 							  $error_lines_ref,
# 							  $warning_lines_ref,
# 							  $where_base )

# function to parse and store a single macro that starts at a given
# offset in a list of text lines, removing it from the list of lines
# sub parse_macro_definition_from_lines_at_offset( $vars_ref, $lines_ref, 
# 						   $macro_start_index,
# 						   $error_lines_ref,
# 						   $warning_lines_ref, 
# 						   $where )

#=======================================================================

# function to parse a constant or define line
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   the line to be parsed, which could be either of the following if match
#      %constant white-space name-string [ white-space [ value-string ]]
#      %define white-space name-string [ white-space [ value-string ]]
#   reference to non-blank fields without comment array
#   reference to error lines array
#   reference to warning lines array
#   where defined string used in warning/error messages
# This function modifies the hash of all flags/values, with new/modified:
#     constants -> reference to %CONSTANT names/values hash
#     defines -> reference to %DEFINE names/values hash
# Returned parameters are:
#   flag which is non-zero if the line had a constant or define definition

sub parse_constant_define_from_line
{
    my( $vars_ref, $line, $fields_without_comment_ref,
	$error_lines_ref, $warning_lines_ref, $where ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};     # save time
#    print "$0: in parse_constant_define_from_line line=<$line>\n"
#	if $debug;

    # find pointers to needed stuff
    my $constants_ref = $$vars_ref{constants};
    my $defines_ref = $$vars_ref{defines};

    # if this line is a %constant or %define, ...
    if ( @$fields_without_comment_ref && 
	 ( $$fields_without_comment_ref[0] eq "%constant" ||
	   $$fields_without_comment_ref[0] eq "%define" ) ) {
#	print "$0: in parse_constant_define_from_line line=<$line>\n"
#	    if $debug;

	# pick out %constant/%define string for error messages
	my @fields = @$fields_without_comment_ref;
	my $type = shift @fields;

	# get the name being assigned to
	if ( ! @fields ) {
	    push @$error_lines_ref, "$type without name $where\n";
	} else {
	    my $name = shift @fields;

	    # if the name is not a legal name, error
	    if ( $name !~ /^[A-Za-z][A-Za-z0-9\_\-]*$/ ) {
		push @$error_lines_ref, 
		"illegal $type name <$name> $where\n";
	    }
	    # if a constant with this name is already defined,
	    # cannot redefine it
	    elsif ( exists $$constants_ref{$name} ) {
		push @$error_lines_ref, 
		"illegal attempt to redefine %constant $name $where";
	    } else {
		# if a %define has been done already, 
		# delete the define and warn of it 
		# before storing the new %constant
		if ( exists $$defines_ref{$name} ) {
#		    push @$warning_lines_ref,
#		    "%define $name being redefined as $type $name $where";
		    delete $$defines_ref{$name};
		}
		# store the new %constant/%define by 
		# taking the non-comment part of the 
		# input line, warning if assigning a 
		# null value since the user probably
		# doesn't want a null value
		my $value;
		if ( @fields ) {
		    $value = join( " ", @fields);
		} else {
		    $value = "";
#		    push @$warning_lines_ref,
#		    "$name $name being assigned a zero length string $where";
		}
		print "$0: in parse_constant_define_from_line storing $type $name=<$value>\n"
		    if $debug;
		if ( $type eq "%constant" ) {
		    $$constants_ref{$name} = "$value";
#		    print_hash("","$0: after constant add constants",
#			       $constants_ref ) if $debug;
		} else {
		    $$defines_ref{$name} = "$value";
#		    print_hash("","$0: after define add defines",
#			       $defines_ref ) if $debug;
		}
	    }
	}
	return 1;
    } else {
	return 0;
    }
}

#=======================================================================

# function to parse constants, defines, and macros definitions from 
# a list of text lines
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to lines array to be parsed and (partially) consumed
#   reference to error lines array to be added to if any errors
#   reference to warning lines array to be added to if any warnings
#   where string without any line number
# This function modifies the hash of all flags/values, with new/modified:
#     constants -> reference to %CONSTANT names/values hash
#     defines -> reference to %DEFINE names/values hash
#     macros -> reference to %MACRO names/references hash
# This function also swallows all of the lines that were used to
# define any constants, defines, or macros .
# Return value is:
#   non-zero if any lines were used and swallowed,
#   zero if no lines were used and swallowed

sub parse_constant_define_macro_definitions_from_lines
{
    my( $vars_ref, $lines_ref, $error_lines_ref,
	$warning_lines_ref, $where_base ) = @_;

    my $debug = ${$$vars_ref{constants}}{debug};     # save time
#    print "$0: in parse_constant_define_macro_definitions_from_lines\n"
#	if $debug;

    # find pointers to needed stuff
    my $constants_ref = $$vars_ref{constants};
    my $defines_ref = $$vars_ref{defines};
    my $macros_ref = $$vars_ref{macros};

    # loop through the input lines
    my $changed = 0;
    my $input_line_number = 0;
    my $line_index = 0;
    my $macro_name;
    my $macro_level = 0;
    while ( $line_index <= $#$lines_ref ) {
	$input_line_number++;
	my $line = $$lines_ref[$line_index];
#	print "$0: in parse_c_d_m_d_from_lines at offset=$line_index line=<$line>\n" 
#	    if $debug;
	my $where = "on line $input_line_number $where_base";

	# find the line without any comments to help out later logic
	my @without_comment_fields = 
	    fields_without_ruleset_comment( $vars_ref, $line );

	# if the line is a %macro start of macro definition,
	# make a new macro, then consume the input lines until
	# we reach the non-nested end of the macro, storing
	# new new macro lines
        if ( parse_macro_definition_from_lines_at_offset( $vars_ref,
							  $lines_ref,
							  $line_index,
							  \@without_comment_fields,
							  $error_lines_ref,
							  $warning_lines_ref,
							  $where_base ) ) {
#	    print_hash("", "after parse_macro_definition_from_lines_at_offset macros",
#		       $$vars_ref{macros} ) if $debug;
	} # end if line was %macro then consuming the macro

	# or if this is an orphaned %endmacro line, error message
        elsif ( @without_comment_fields && 
		$line =~ /^\%/ &&
		$without_comment_fields[0] eq "%endmacro" ) {
	    push @$error_lines_ref, "%endmacro without %macro $where";
	    $lines_index++; # swallow/remove %endmacro line from lines
	}

	# or if this line is a %constant or %define store it
	elsif ( parse_constant_define_from_line ( $vars_ref,
						  $$lines_ref[$line_index],
						  \@without_comment_fields,
						  $error_lines_ref,
						  $warning_lines_ref,
						  $where ) ) {
	    $line_index++; # swallow/remove %constant/%define line
	}

	# or this is a line we ignore, bump past it
	else {
	    $line_index++; # swallow/remove unused line from lines
	}

    # end of loop for each line
    }
}

#=======================================================================

# function to parse and store a single macro that starts at a given 
# offset in a list of text lines, removing it from the list of lines
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to lines array to be parsed and (partially) consumed
#   offset into the lines array that has the starting %macro line
#   reference to non-blank fields without comment
#   reference to error lines array to be added to if any errors
#   reference to warning lines array to be added to if any warnings
#   where string without any line number
# This function modifies the hash of all flags/values, with new/modified:
#     constants -> reference to %CONSTANT names/values hash
#     defines -> reference to %DEFINE names/values hash
#     macros -> reference to %MACRO names/references hash
# This function also swallows all of the lines that were used to
# define any constants, defines, or macros .
# Returned parameters are:
#   This function returns non-zero if the line started with a 
#   %macro field, or returns zero if not.

sub parse_macro_definition_from_lines_at_offset
{
    my( $vars_ref, $lines_ref, $macro_start_index,
	$fields_without_comment_ref,
	$error_lines_ref, $warning_lines_ref, $where ) = @_;

    my $debug = ${$$vars_ref{constants}}{debug};     # save time
#    if ( $debug ) {
#	my $string = ( $macro_start_index < $#$lines_ref ) ? 
#	    "line=<$$lines_ref[$macro_start_index]>\n" : 
#		"last line index=$#$lines_ref\n";
#	print "$0: in parse_macro_definition_from_lines_at_offset offset=$macro_start_index $string";
#    }

    return 0 if ( $macro_start_index > $#$lines_ref );

    # find pointers to needed stuff
    my $macros_ref = $$vars_ref{macros};

    my $line_index = $macro_start_index;
    my $line = $$lines_ref[$line_index];
    my @fields = @$fields_without_comment_ref;

    # if the line is a %macro start of macro definition,
    # make a new macro, then consume the input lines until
    # we reach the non-nested end of the macro, storing
    # new new macro lines
    if ( @fields &&
	 $line =~ /^\%/ &&
	 $fields[0] eq "%macro" ) {

	# init index past where macro ends
	my $macro_past_index = $line_index + 1;

	# if no macro name, error
	if ( @fields < 2 ) {
	    push @$error_lines_ref, 
	    "no macro name or arguments $where\n";
	} else {

	    # get the name of the macro
	    shift @fields;
	    $macro_name = shift @fields;
	    $macro_name =~ s/^\%//;
#	    print "$0: starting definition of macro $macro_name $where\n"
#		if $debug;

	    # see if this macro already exists and if so,
	    # get rid of the old definition and warn about
	    # redefining the macro
	    if ( exists $$macros_ref{$macro_name} ) {
		push @$warning_lines_ref, 
		"macro $macro_name being redefined $where\n";
		delete $$macros_ref{$macro_name};
	    }

	    # initialize a new macro structure and point 
	    # macros hash entry for it's name to it
	    my %macro;
	    my $macro_ref = \%macro;
	    $$macros_ref{$macro_name} = $macro_ref;
	    $$macro_ref{name} = $macro_name;
	    my @macro_arguments = @fields;
	    my @macro_lines = ();
	    $$macro_ref{arguments} = \@macro_arguments;
	    $$macro_ref{lines} = \@macro_lines;
	    $$macro_ref{defined_where} = $where;
#	    print_hash("", "initialized macro $macro_name", $macro_ref)
#		if $debug;
	    $macro_level = 1;

	    # consume and store macro lines until nested %endmacro
	    while ( $macro_level &&
		    $macro_past_index <= $#$lines_ref ) {
		$input_line_number++; # update original input index

		# get the line
		my $macro_line = $$lines_ref[$macro_past_index];
		$macro_past_index++; # update index past where we're looking

		# if the line is a %endmacro end of macro definition,
		# decrease the macro level count if we are in nested
		# macro definitions, or close the current macro if not
		my @macro_line_fields = 
		    fields_without_ruleset_comment( $vars_ref, $macro_line );
		if ( @macro_line_fields &&
		     $macro_line_fields[0] eq "%endmacro" ) {
		    $macro_level--;
		    if ( ! $macro_level ) {
#			print "$0: ending definition of macro $macro_name $where\n"
#			    if $debug;
#			print_hash("", "finished macro $macro_name", $macro_ref)
#			    if $debug;
		    }
		} else {
		    push @{$$macro_ref{lines}}, $macro_line;
#                        print "$0: macro $macro_name now has " . 
#                        scalar @{$$macro_ref{lines}} . " lines\n"
#                           if $debug;
		}
	    }

	    # if the macro wasn't finished in these lines, error
	    push @$error_lines_ref, "macro $macro_name has no %endmacro"
		if $macro_level;
	}

	# cut the macro definition lines from the input lines
	my $macro_size = $macro_past_index - $macro_start_index - 2;
#	print "$0: after storing $macro_size lines in macro $macro_name removing " .
#		($macro_size + 2) .
#		    " lines at offset $macro_start_index\n"
#			if $debug;
	splice @$lines_ref, $macro_start_index, $macro_size + 2;
#	print "$0: after removing " . ($macro_size + 2) . 
#	    " lines input has " . scalar @$lines_ref . 
#	    " lines total\n" if $debug;

	return 1;
    } else {
	return 0;
    }
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;
